<?php

namespace Drupal\gbif2_entity\Plugin\ExternalEntities\StorageClient;

use Drupal\Core\Entity\EntityFieldManagerInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Logger\LoggerChannelFactoryInterface;
use Drupal\Core\Messenger\MessengerTrait;
use Drupal\Core\Plugin\PluginFormInterface;
use Drupal\Core\StringTranslation\TranslationInterface;
use Drupal\Core\Utility\Token;
use Drupal\external_entities\Entity\ExternalEntityInterface;
use Drupal\external_entities\Plugin\PluginFormTrait;
use Drupal\external_entities\StorageClient\StorageClientBase;
use ResTelae\Gbif\Occurrences;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * External entities storage client based on the GBIF REST API.
 *
 * @StorageClient(
 *   id = "gbif",
 *   label = @Translation("GBIF"),
 *   description = @Translation("Retrieves GBIF records")
 * )
 */
class Gbif extends StorageClientBase implements PluginFormInterface {

  use PluginFormTrait;
  use MessengerTrait;

  /**
   * GBIF occurrences service, provided by the PHP GBIF library.
   *
   * @var \ResTelae\Gbif\Occurrences
   */
  protected $occ;

  /**
   * Constructs a Gbif object.
   *
   * @param array $configuration
   *   A configuration array containing information about the plugin instance.
   * @param string $plugin_id
   *   The plugin_id for the plugin instance.
   * @param mixed $plugin_definition
   *   The plugin implementation definition.
   * @param \Drupal\Core\StringTranslation\TranslationInterface $string_translation
   *   The string translation service.
   * @param \Drupal\Core\Logger\LoggerChannelFactoryInterface $logger_factory
   *   The logger channel factory.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity type manager.
   * @param \Drupal\Core\Entity\EntityFieldManagerInterface $entity_field_manager
   *   The entity field manager service.
   * @param \Drupal\Core\Utility\Token $token_service
   *   The token service.
   * @param \ResTelae\Gbif\Occurrences $occ
   *   GBIF occurrences.
   */
  public function __construct(
    array $configuration,
    string $plugin_id,
    $plugin_definition,
    TranslationInterface $string_translation,
    LoggerChannelFactoryInterface $logger_factory,
    EntityTypeManagerInterface $entity_type_manager,
    EntityFieldManagerInterface $entity_field_manager,
    Token $token_service,
    Occurrences $occ,
  ) {
    parent::__construct(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $string_translation,
      $logger_factory,
      $entity_type_manager,
      $entity_field_manager,
      $token_service
    );
    $this->occ = $occ;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(
    ContainerInterface $container,
    array $configuration,
    $plugin_id,
    $plugin_definition,
  ) {
    return new static(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $container->get('string_translation'),
      $container->get('logger.factory'),
      $container->get('entity_type.manager'),
      $container->get('entity_field.manager'),
      $container->get('token'),
      $container->get('gbif.occurrence')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function defaultConfiguration() {
    return [];
  }

  /**
   * {@inheritdoc}
   */
  public function buildConfigurationForm(
    array $form,
    FormStateInterface $form_state,
  ) {
    $form['info'] = [
      '#type' => 'item',
      '#markup' => $this->t(
        'There is no configuration for the GBIF storage client. You can map the "key" field as entity identifier field, "occurrenceID" as UUID and "scientificName" as title. See <a href="https://techdocs.gbif.org/en/openapi/v1/occurrence#/Occurrences/getOccurrenceById">GBIF Occurrence API</a> for available field list.'
      ),
    ];
    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function load(string|int $id) :array|null {
    return $this->occ->get($id);
  }

  /**
   * {@inheritdoc}
   */
  public function loadMultiple(?array $ids = NULL) :array {
    $data = [];

    if (!empty($ids) && is_array($ids)) {
      foreach ($ids as $id) {
        $data[$id] = $this->load($id);
      }
    }

    return $data;
  }

  /**
   * {@inheritdoc}
   *
   * @note GBIF API does not allow sorting.
   */
  public function querySource(
    array $parameters = [],
    array $sorts = [],
    ?int $start = NULL,
    ?int $length = NULL,
  ) :array {
    $results = $this->search($parameters, $start, $length);

    return $results['results'] ?? [];
  }

  /**
   * {@inheritdoc}
   */
  public function countQuerySource(array $parameters = []) :int {
    $results = $this->search($parameters, 0, 1);
    return $results['count'] ?? 0;
  }

  /**
   * Search from GBIF.
   *
   * @param array $parameters
   *   (optional) Array of parameters, each value is an array with the following
   *   key-value pairs:
   *     - field: the field name the parameter applies to
   *     - value: the value of the parameter
   *     - operator: the operator of how the parameter should be applied.
   * @param int|null $start
   *   (optional) The first item to return.
   * @param int|null $length
   *   (optional) The number of items to return.
   *
   * @return array
   *   Results retrieved from GBIF.
   */
  protected function search(
    array $parameters = [],
    $start = NULL,
    $length = NULL,
  ) :array {
    $params = [
      'limit' => $length,
      'offset' => $start,
    ];

    foreach ($parameters as $parameter) {
      // We do not bother with fancy operators.
      // @see \Drupal\gbif2_views\Plugin\views\filter\GbifNumericFilter
      $params[$parameter['field']] = $parameter['value'];
    }

    return $this->occ->search($params);
  }

  /**
   * {@inheritdoc}
   */
  public function save(ExternalEntityInterface $entity) :int {
    // Not currently supported.
    return 0;
  }

  /**
   * {@inheritdoc}
   */
  public function delete(ExternalEntityInterface $entity) {
    // Not currently supported.
  }

  /**
   * {@inheritdoc}
   */
  public function transliterateDrupalFilters(
    array $parameters,
    array $context = [],
  ) :array {
    if (1 <= $this->getDebugLevel()) {
      $this->logger->debug(
        "Gbif::transliterateDrupalFilters():\n@parameters",
        [
          '@parameters' => print_r($parameters, TRUE),
        ]
      );
    }
    // @todo Improve this part to better reflect what can be filtered.
    $drupal_filters = [];
    $source_filters = [];
    foreach ($parameters as $parameter) {
      if (!isset($parameter['field'])
        || !isset($parameter['value'])
        || !is_string($parameter['field'])
        || (FALSE !== strpos($parameter['field'], '.'))
      ) {
        // Pass complex filters to Drupal.
        $drupal_filters[] = $parameter;
        if (1 <= $this->getDebugLevel()) {
          $this->logger->debug(
            "Complex filter passed to Drupal: @filter",
            [
              '@filter' => print_r($parameter, TRUE),
            ]
          );
        }
      }
      else {
        // Try to handle simple filters on the GBIF2-side.
        $field_mapper = $this->externalEntityType->getFieldMapper($parameter['field']);
        if ($field_mapper) {
          $source_field = $field_mapper->getMappedSourceFieldName();
        }
        // We drop non-mapped field properties and ignore operators.
        if (!empty($source_field)) {
          $source_filters[] = [
            'field' => $source_field,
            'value' => $parameter['value'],
            'operator' => '=',
          ];
        }
        elseif (1 <= $this->getDebugLevel()) {
          $this->logger->debug(
            "Skipped filter for field '@field': no mapping",
            [
              '@field' => $parameter['field'],
            ]
          );
        }
      }
    }

    if (1 <= $this->getDebugLevel()) {
      $this->logger->debug(
        "Transliteration done. Returning:\n@filters",
        [
          '@filters' => print_r(
            ['source' => $source_filters, 'drupal' => $drupal_filters],
            TRUE
          ),
        ]
      );
    }

    return ['source' => $source_filters, 'drupal' => $drupal_filters];
  }

}
