<?php

namespace Drupal\gbif2_views\Plugin\views\query;

use Drupal\views\Plugin\views\query\QueryPluginBase;
use Drupal\views\ResultRow;
use Drupal\views\ViewExecutable;
use ResTelae\Gbif\Occurrences;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Views query class using GBIF as the data source.
 *
 * @ViewsQuery(
 *   id = "gbif_occurrence",
 *   title = @Translation("Gbif Occurrences"),
 *   help = @Translation("Query against the GBIF API.")
 * )
 */
class GbifOccurrence extends QueryPluginBase {

  /**
   * GBIF occurrences service, provided by the PHP GBIF library.
   *
   * @var \ResTelae\Gbif\Occurrences
   */
  protected $occ;

  /**
   * Fitbit constructor.
   *
   * @param array $configuration
   *   Plugin configuration.
   * @param string $plugin_id
   *   Plugin ID.
   * @param mixed $plugin_definition
   *   Plugin definition.
   * @param \ResTelae\Gbif\Occurrences $occ
   *   GBIF occurrences services, provided by PHP GBIF.
   */
  public function __construct(array $configuration, $plugin_id, $plugin_definition, Occurrences $occ) {
    parent::__construct($configuration, $plugin_id, $plugin_definition);
    $this->occ = $occ;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    return new static(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $container->get('gbif.occurrence')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function ensureTable($table, $relationship = NULL) {
    // Not an SQL-query backend.
    return '';
  }

  /**
   * {@inheritdoc}
   */
  public function addField($table, $field, $alias = '', $params = []) {
    // We don't really care, not an SQL-query backend, remember?
    return $field;
  }

  /**
   * {@inheritdoc}
   *
   * We do not really use this, unless $field value is 'key'.
   */
  public function addWhere($group, $field, $value = NULL, $operator = NULL) {
    // Ensure all variants of 0 are actually 0. Thus '', 0 and NULL are all
    // the default group.
    if (empty($group)) {
      $group = 0;
    }
    // Check for a group.
    if (!isset($this->where[$group])) {
      $this->setWhereGroup('AND', $group);
    }
    $this->where[$group]['conditions'][] = [
      'field' => $field,
      'value' => $value,
      'operator' => $operator,
    ];
  }

  /**
   * Builds the necessary info to execute the query.
   */
  public function build(ViewExecutable $view) {
    $view->build_info['query'] = [];

    // Filters.
    foreach ($view->filter as $filter) {
      $view->build_info['query'] += $filter->generate();
    }

    // Where conditions (only for key).
    if (isset($this->where[0]['conditions'])) {
      foreach ($this->where[0]['conditions'] as $condition) {
        $operators = ['=', 'IN'];
        if ($condition['field'] == 'key' && in_array($condition['operator'], $operators)) {
          $view->build_info['query']['key'] = $condition['value'];
        }
      }
    }

    // Pager.
    $view->initPager();
    if ($view->pager->usePager()) {
      $items_per_page = $view->pager->getItemsPerPage();
      $view->build_info['query'] += [
        'limit' => $items_per_page,
        'offset' => $view->pager->getCurrentPage() * $items_per_page,
      ];
    }
  }

  /**
   * {@inheritdoc}
   */
  public function execute(ViewExecutable $view) {
    $start = microtime(TRUE);
    $index = 0;
    $results = $this->getResults($view);

    foreach ($results['results'] as $data) {
      $row = array_intersect_key($data, $view->field);
      // Always include ID no matter what.
      $row['key'] = $data['key'];
      $row['index'] = $index++;
      $view->result[] = new ResultRow($row);
    }

    $view->total_rows = $view->pager->total_items = $results['count'];
    $view->pager->updatePageInfo();
    $view->execute_time = microtime(TRUE) - $start;
  }

  /**
   * Get view results.
   *
   * @param \Drupal\views\ViewExecutable $view
   *   The view resuls will be retrieved for.
   *
   * @return array
   *   An array with two keys:
   *    - results contains an array of GBIF occurrences,
   *    - count the number of results.
   */
  protected function getResults(ViewExecutable $view) {
    $results = [];

    if (isset($view->build_info['query']['key'])) {
      // If key is specified, GBIF search API is not called as it does not allow
      // to specify this as a parameter. Instead, we directly retrieve
      // occurrences. This behaviour takes precedence.
      $keys = $view->build_info['query']['key'];
      $keys = is_array($keys) ? $keys : [$keys];
      foreach ($keys as $key) {
        $results['results'][] = $this->occ->get($key);
      }
      $results['count'] = count($keys);
    }
    else {
      $results = $this->occ->search($view->build_info['query']);
    }

    return $results;
  }

}
